"""TexSoup transforms a LaTeX document into a complex tree of various Python
objects, but all objects fall into one of the following three categories:

``TexNode``, ``TexExpr`` (environments and commands), and ``TexGroup`` s.
"""

import itertools
import re
from .utils import CharToLineOffset, Token, TC, to_list

__all__ = ['TexNode', 'TexCmd', 'TexEnv', 'TexGroup', 'BracketGroup',
           'BraceGroup', 'TexArgs', 'TexText', 'TexMathEnv',
           'TexDisplayMathEnv', 'TexNamedEnv', 'TexMathModeEnv',
           'TexDisplayMathModeEnv']


#############
# Interface #
#############


class TexNode(object):
    r"""A tree node representing an expression in the LaTeX document.

    Every node in the parse tree is a ``TexNode``, equipped with navigation,
    search, and modification utilities. To navigate the parse tree, use
    abstractions such as ``children`` and ``descendant``. To access content in
    the parse tree, use abstractions such as ``contents``, ``text``, ``string``
    , and ``args``.

    Note that the LaTeX parse tree is largely shallow: only environments such
    as ``itemize`` or ``enumerate`` have children and thus descendants. Typical
    LaTeX expressions such as ``\section`` have *arguments* but not children.
    """

    def __init__(self, expr, src=None):
        """Creates TexNode object.

        :param TexExpr expr: a LaTeX expression, either a singleton
            command or an environment containing other commands
        :param str src: LaTeX source string
        """
        assert isinstance(expr, TexExpr), \
            'Expression given to node must be a valid TexExpr'
        super().__init__()
        self.expr = expr
        self.parent = None
        if src is not None:
            self.char_to_line = CharToLineOffset(src)
        else:
            self.char_to_line = None

    #################
    # MAGIC METHODS #
    #################

    def __contains__(self, other):
        """Use custom containment checker where applicable (TexText, for ex)"""
        if hasattr(self.expr, '__contains__'):
            return other in self.expr
        return other in iter(self)

    def __getattr__(self, attr, default=None):
        """Convert all invalid attributes into basic find operation."""
        return self.find(attr) or default

    def __getitem__(self, item):
        return list(self.contents)[item]

    def __iter__(self):
        """
        >>> node = TexNode(TexNamedEnv('lstlisting', ('hai', 'there')))
        >>> list(node)
        ['hai', 'there']
        """
        return iter(self.contents)

    def __match__(self, name=None, attrs=()):
        r"""Check if given attributes match current object

        >>> from TexSoup import TexSoup
        >>> soup = TexSoup(r'\ref{hello}\ref{hello}\ref{hello}\ref{nono}')
        >>> soup.count(r'\ref{hello}')
        3
        """
        return self.expr.__match__(name, attrs)

    def __repr__(self):
        """Interpreter representation."""
        return str(self)

    def __str__(self):
        """Stringified command."""
        return str(self.expr)

    ##############
    # PROPERTIES #
    ##############

    @property
    @to_list
    def all(self):
        r"""Returns all content in this node, regardless of whitespace or
        not. This includes all LaTeX needed to reconstruct the original source.

        >>> from TexSoup import TexSoup
        >>> soup = TexSoup(r'''
        ... \newcommand{reverseconcat}[3]{#3#2#1}
        ... ''')
        >>> alls = soup.all
        >>> alls[0]
        <BLANKLINE>
        <BLANKLINE>
        >>> alls[1]
        \newcommand{reverseconcat}[3]{#3#2#1}
        """
        for child in self.expr.all:
            assert isinstance(child, TexExpr)
            node = TexNode(child)
            node.parent = self
            yield node

    @property
    def args(self):
        r"""Arguments for this node. Note that this argument is settable.

        :rtype: TexArgs

        >>> from TexSoup import TexSoup
        >>> soup = TexSoup(r'''\newcommand{reverseconcat}[3]{#3#2#1}''')
        >>> soup.newcommand.args
        [BraceGroup('reverseconcat'), BracketGroup('3'), BraceGroup('#3#2#1')]
        >>> soup.newcommand.args = soup.newcommand.args[:2]
        >>> soup.newcommand
        \newcommand{reverseconcat}[3]
        """
        return self.expr.args

    @args.setter
    def args(self, args):
        assert isinstance(args, TexArgs), "`args` must be of type `TexArgs`"
        self.expr.args = args

    @property
    @to_list
    def children(self):
        r"""Immediate children of this TeX element that are valid TeX objects.

        This is equivalent to contents, excluding text elements and keeping
        only Tex expressions.

        :return: generator of all children
        :rtype: Iterator[TexExpr]

        >>> from TexSoup import TexSoup
        >>> soup = TexSoup(r'''
        ... \begin{itemize}
        ...     Random text!
        ...     \item Hello
        ... \end{itemize}''')
        >>> soup.itemize.children[0]
        \item Hello
        <BLANKLINE>
        """
        for child in self.expr.children:
            node = TexNode(child)
            node.parent = self
            yield node

    @property
    @to_list
    def contents(self):
        r"""Any non-whitespace contents inside of this TeX element.

        :return: generator of all nodes, tokens, and strings
        :rtype: Iterator[Union[TexNode,str]]

        >>> from TexSoup import TexSoup
        >>> soup = TexSoup(r'''
        ... \begin{itemize}
        ...     Random text!
        ...     \item Hello
        ... \end{itemize}''')
        >>> contents = soup.itemize.contents
        >>> contents[0]
        '\n    Random text!\n    '
        >>> contents[1]
        \item Hello
        <BLANKLINE>
        """
        for child in self.expr.contents:
            if isinstance(child, TexExpr):
                node = TexNode(child)
                node.parent = self
                yield node
            else:
                yield child

    @contents.setter
    def contents(self, contents):
        self.expr.contents = contents

    @property
    def descendants(self):
        r"""Returns all descendants for this TeX element.

        >>> from TexSoup import TexSoup
        >>> soup = TexSoup(r'''
        ... \begin{itemize}
        ...     \begin{itemize}
        ...         \item Nested
        ...     \end{itemize}
        ... \end{itemize}''')
        >>> descendants = list(soup.itemize.descendants)
        >>> descendants[1]
        \item Nested
        <BLANKLINE>
        """
        return self.__descendants()

    @property
    def name(self):
        r"""Name of the expression. Used for search functions.

        :rtype: str

        >>> from TexSoup import TexSoup
        >>> soup = TexSoup(r'''\textbf{Hello}''')
        >>> soup.textbf.name
        'textbf'
        >>> soup.textbf.name = 'textit'
        >>> soup.textit
        \textit{Hello}
        """
        return self.expr.name

    @name.setter
    def name(self, name):
        self.expr.name = name

    @property
    def string(self):
        r"""This is valid if and only if

        1. the expression is a :class:`.TexCmd` AND has only one argument OR
        2. the expression is a :class:`.TexEnv` AND has only one TexText child

        :rtype: Union[None,str]

        >>> from TexSoup import TexSoup
        >>> soup = TexSoup(r'''\textbf{Hello}''')
        >>> soup.textbf.string
        'Hello'
        >>> soup.textbf.string = 'Hello World'
        >>> soup.textbf.string
        'Hello World'
        >>> soup.textbf
        \textbf{Hello World}
        >>> soup = TexSoup(r'''\begin{equation}1+1\end{equation}''')
        >>> soup.equation.string
        '1+1'
        >>> soup.equation.string = '2+2'
        >>> soup.equation.string
        '2+2'
        """
        if isinstance(self.expr, TexCmd):
            assert len(self.expr.args) == 1, \
                '.string is only valid for commands with one argument'
            return self.expr.args[0].string

        contents = list(self.contents)
        if isinstance(self.expr, TexEnv):
            assert len(contents) == 1 and \
                isinstance(contents[0], (TexText, str)), \
                '.string is only valid for environments with only text content'
            return contents[0]

    @string.setter
    def string(self, string):
        if isinstance(self.expr, TexCmd):
            assert len(self.expr.args) == 1, \
                '.string is only valid for commands with one argument'
            self.expr.args[0].string = string

        contents = list(self.contents)
        if isinstance(self.expr, TexEnv):
            assert len(contents) == 1 and \
                isinstance(contents[0], (TexText, str)), \
                '.string is only valid for environments with only text content'
            self.contents = [string]

    @property
    def position(self):
        r"""Position of first character in expression, in original source.

        Note this position is NOT updated as the parsed tree is modified.
        """
        return self.expr.position

    @property
    @to_list
    def text(self):
        r"""All text in descendant nodes.

        This is equivalent to contents, keeping text elements and excluding
        Tex expressions.

        >>> from TexSoup import TexSoup
        >>> soup = TexSoup(r'''
        ... \begin{itemize}
        ...     \begin{itemize}
        ...         \item Nested
        ...     \end{itemize}
        ... \end{itemize}''')
        >>> soup.text[0]
        ' Nested\n    '
        """
        for descendant in self.contents:
            if isinstance(descendant, (TexText, Token)):
                yield descendant
            elif hasattr(descendant, 'text'):
                yield from descendant.text

    ##################
    # PUBLIC METHODS #
    ##################

    def append(self, *nodes):
        r"""Add node(s) to this node's list of children.

        :param TexNode nodes: List of nodes to add

        >>> from TexSoup import TexSoup
        >>> soup = TexSoup(r'''
        ... \begin{itemize}
        ...     \item Hello
        ... \end{itemize}
        ... \section{Hey}
        ... \textit{Willy}''')
        >>> soup.section
        \section{Hey}
        >>> soup.section.append(soup.textit)  # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        TypeError: ...
        >>> soup.section
        \section{Hey}
        >>> soup.itemize.append('    ', soup.item)
        >>> soup.itemize
        \begin{itemize}
            \item Hello
            \item Hello
        \end{itemize}
        """
        self.expr.append(*nodes)

    def insert(self, i, *nodes):
        r"""Add node(s) to this node's list of children, at position i.

        :param int i: Position to add nodes to
        :param TexNode nodes: List of nodes to add

        >>> from TexSoup import TexSoup
        >>> soup = TexSoup(r'''
        ... \begin{itemize}
        ...     \item Hello
        ...     \item Bye
        ... \end{itemize}''')
        >>> item = soup.item.copy()
        >>> soup.item.delete()
        >>> soup.itemize.insert(1, item)
        >>> soup.itemize
        \begin{itemize}
            \item Hello
            \item Bye
        \end{itemize}
        >>> item.parent.name == soup.itemize.name
        True
        """
        assert isinstance(i, int), (
            'Provided index "{}" is not an integer! Did you switch your '
            'arguments? The first argument to `insert` is the '
            'index.'.format(i))
        for node in nodes:
            if not isinstance(node, TexNode):
                continue
            assert not node.parent, (
                'Inserted node should not already have parent. Call `.copy()` '
                'on node to fix.'
            )
            node.parent = self

        self.expr.insert(i, *nodes)

    def char_pos_to_line(self, char_pos):
        r"""Map position in the original string to parsed LaTeX position.

        :param int char_pos: Character position in the original string
        :return: (line number, index of character in line)
        :rtype: Tuple[int, int]

        >>> from TexSoup import TexSoup
        >>> soup = TexSoup(r'''
        ... \section{Hey}
        ... \textbf{Silly}
        ... \textit{Willy}''')
        >>> soup.char_pos_to_line(10)
        (1, 9)
        >>> soup.char_pos_to_line(20)
        (2, 5)
        """
        assert self.char_to_line is not None, (
            'CharToLineOffset is not initialized. Pass src to TexNode '
            'constructor')
        return self.char_to_line(char_pos)

    def copy(self):
        r"""Create another copy of the current node.

        >>> from TexSoup import TexSoup
        >>> soup = TexSoup(r'''
        ... \section{Hey}
        ... \textit{Silly}
        ... \textit{Willy}''')
        >>> s = soup.section.copy()
        >>> s.parent is None
        True
        """
        return TexNode(self.expr)

    def count(self, name=None, **attrs):
        r"""Number of descendants matching criteria.

        :param Union[None,str] name: name of LaTeX expression
        :param attrs: LaTeX expression attributes, such as item text.
        :return: number of matching expressions
        :rtype: int

        >>> from TexSoup import TexSoup
        >>> soup = TexSoup(r'''
        ... \section{Hey}
        ... \textit{Silly}
        ... \textit{Willy}''')
        >>> soup.count('section')
        1
        >>> soup.count('textit')
        2
        """
        return len(list(self.find_all(name, **attrs)))

    def delete(self):
        r"""Delete this node from the parse tree.

        Where applicable, this will remove all descendants of this node from
        the parse tree.

        >>> from TexSoup import TexSoup
        >>> soup = TexSoup(r'''
        ... \textit{\color{blue}{Silly}}\textit{keep me!}''')
        >>> soup.textit.color.delete()
        >>> soup
        <BLANKLINE>
        \textit{}\textit{keep me!}
        >>> soup.textit.delete()
        >>> soup
        <BLANKLINE>
        \textit{keep me!}
        """

        # TODO: needs better abstraction for supports contents
        parent = self.parent
        if parent.expr._supports_contents():
            parent.remove(self)
            return

        # TODO: needs abstraction for removing from arg
        for arg in parent.args:
            if self.expr in arg.contents:
                arg._contents.remove(self.expr)

    def find(self, name=None, **attrs):
        r"""First descendant node matching criteria.

        Returns None if no descendant node found.

        :return: descendant node matching criteria
        :rtype: Union[None,TexExpr]

        >>> from TexSoup import TexSoup
        >>> soup = TexSoup(r'''
        ... \section{Ooo}
        ... \textit{eee}
        ... \textit{ooo}''')
        >>> soup.find('textit')
        \textit{eee}
        >>> soup.find('textbf')
        """
        try:
            return self.find_all(name, **attrs)[0]
        except IndexError:
            return None

    @to_list
    def find_all(self, name=None, **attrs):
        r"""Return all descendant nodes matching criteria.

        :param Union[None,str,list] name: name of LaTeX expression
        :param attrs: LaTeX expression attributes, such as item text.
        :return: All descendant nodes matching criteria
        :rtype: Iterator[TexNode]

        If `name` is a list of `str`'s, any matching section will be matched.

        >>> from TexSoup import TexSoup
        >>> soup = TexSoup(r'''
        ... \section{Ooo}
        ... \textit{eee}
        ... \textit{ooo}''')
        >>> gen = soup.find_all('textit')
        >>> gen[0]
        \textit{eee}
        >>> gen[1]
        \textit{ooo}
        >>> soup.find_all('textbf')[0]
        Traceback (most recent call last):
        ...
        IndexError: list index out of range
        """
        for descendant in self.__descendants():
            if hasattr(descendant, '__match__') and \
                    descendant.__match__(name, attrs):
                yield descendant

    def remove(self, node):
        r"""Remove a node from this node's list of contents.

        :param TexExpr node: Node to remove

        >>> from TexSoup import TexSoup
        >>> soup = TexSoup(r'''
        ... \begin{itemize}
        ...     \item Hello
        ...     \item Bye
        ... \end{itemize}''')
        >>> soup.itemize.remove(soup.item)
        >>> soup.itemize
        \begin{itemize}
            \item Bye
        \end{itemize}
        """
        self.expr.remove(node.expr)

    def replace_with(self, *nodes):
        r"""Replace this node in the parse tree with the provided node(s).

        :param TexNode nodes: List of nodes to subtitute in

        >>> from TexSoup import TexSoup
        >>> soup = TexSoup(r'''
        ... \begin{itemize}
        ...     \item Hello
        ...     \item Bye
        ... \end{itemize}''')
        >>> items = list(soup.find_all('item'))
        >>> bye = items[1]
        >>> soup.item.replace_with(bye)
        >>> soup.itemize
        \begin{itemize}
            \item Bye
        \item Bye
        \end{itemize}
        """
        self.parent.replace(self, *nodes)

    def replace(self, child, *nodes):
        r"""Replace provided node with node(s).

        :param TexNode child: Child node to replace
        :param TexNode nodes: List of nodes to subtitute in

        >>> from TexSoup import TexSoup
        >>> soup = TexSoup(r'''
        ... \begin{itemize}
        ...     \item Hello
        ...     \item Bye
        ... \end{itemize}''')
        >>> items = list(soup.find_all('item'))
        >>> bye = items[1]
        >>> soup.itemize.replace(soup.item, bye)
        >>> soup.itemize
        \begin{itemize}
            \item Bye
        \item Bye
        \end{itemize}
        """
        for arg in self.expr.args:
            if child.expr in arg._contents:
                arg.insert(arg.remove(child.expr), *nodes)
                return
        self.expr.insert(
            self.expr.remove(child.expr),
            *nodes)


    def search_regex(self, pattern):
        for node in self.text:
            for match in re.finditer(pattern, node):
                body = match.group()  # group() returns the full match
                start = match.start()
                yield Token(body, node.position + start)

    def __descendants(self):
        """Implementation for descendants, hacky workaround for __getattr__
        issues."""
        return itertools.chain(self.contents,
                               *[c.descendants for c in self.children])


###############
# Expressions #
###############


class TexExpr(object):
    """General abstraction for a TeX expression.

    An expression may be a command or an environment and is identified
    by a name, arguments, and place in the parse tree. This is an
    abstract and is not directly instantiated.
    """

    def __init__(self, name, contents=(), args=(), preserve_whitespace=False,
                 position=-1):
        """Initialize a tex expression.

        :param str name: name of environment
        :param iterable contents: list of contents
        :param iterable args: list of Tex Arguments
        :param bool preserve_whitespace: If false, elements containing only
            whitespace will be removed from contents.
        :param int position: position of first character in original source
        """
        self.name = name.strip()  # TODO: should not ever have space
        self.args = TexArgs(args)
        self.parent = None
        self._contents = list(contents) or []
        self.preserve_whitespace = preserve_whitespace
        self.position = position

        for content in contents:
            if isinstance(content, (TexEnv, TexCmd)):
                content.parent = self

    #################
    # MAGIC METHODS #
    #################

    def __eq__(self, other):
        """Check if two expressions are equal. This is useful when defining
        data structures over TexExprs.

        >>> exprs = [
        ...     TexExpr('cake', ['flour', 'taro']),
        ...     TexExpr('corgi', ['derp', 'collar', 'drool', 'sass'])
        ... ]
        >>> exprs[0] in exprs
        True
        >>> TexExpr('cake', ['flour', 'taro']) in exprs
        True
        """
        return str(other) == str(self)

    def __match__(self, name=None, attrs=()):
        """Check if given attributes match current object."""
        # TODO: this should re-parse the name, instead of hardcoding here
        if '{' in name or '[' in name:
            return str(self) == name
        if isinstance(name, list):
            node_name = getattr(self, 'name')
            if node_name not in name:
                return False
        else:
            attrs['name'] = name
        for k, v in attrs.items():
            if getattr(self, k) != v:
                return False
        return True

    def __repr__(self):
        if not self.args:
            return "TexExpr('%s', %s)" % (self.name, repr(self._contents))
        return "TexExpr('%s', %s, %s)" % (
            self.name, repr(self._contents), repr(self.args))

    ##############
    # PROPERTIES #
    ##############

    @property
    @to_list
    def all(self):
        r"""Returns all content in this expression, regardless of whitespace or
        not. This includes all LaTeX needed to reconstruct the original source.

        >>> expr1 = TexExpr('textbf', ('\n', 'hi'))
        >>> expr2 = TexExpr('textbf', ('\n', 'hi'), preserve_whitespace=True)
        >>> list(expr1.all) == list(expr2.all)
        True
        """
        for arg in self.args:
            for expr in arg.contents:
                yield expr
        for content in self._contents:
            yield content

    @property
    @to_list
    def children(self):
        return filter(lambda x: isinstance(x, (TexEnv, TexCmd)), self.contents)

    @property
    @to_list
    def contents(self):
        r"""Returns all contents in this expression.

        Optionally includes whitespace if set when node was created.

        >>> expr1 = TexExpr('textbf', ('\n', 'hi'))
        >>> list(expr1.contents)
        ['hi']
        >>> expr2 = TexExpr('textbf', ('\n', 'hi'), preserve_whitespace=True)
        >>> list(expr2.contents)
        ['\n', 'hi']
        >>> expr = TexExpr('textbf', ('\n', 'hi'))
        >>> expr.contents = ('hehe', '👻')
        >>> list(expr.contents)
        ['hehe', '👻']
        >>> expr.contents = 35  #doctest:+ELLIPSIS
        Traceback (most recent call last):
            ...
        TypeError: ...
        """
        for content in self.all:
            if isinstance(content, TexText):
                content = content._text
            is_whitespace = isinstance(content, str) and content.isspace()
            if not is_whitespace or self.preserve_whitespace:
                yield content

    @contents.setter
    def contents(self, contents):
        if not isinstance(contents, (list, tuple)) or not all(
                isinstance(content, (str, TexExpr)) for content in contents):
            raise TypeError(
                '.contents value "%s" must be a list or tuple of strings or '
                'TexExprs' % contents)
        _contents = [TexText(c) if isinstance(c, str) else c for c in contents]
        self._contents = _contents

    @property
    def string(self):
        """All contents stringified. A convenience property

        >>> expr = TexExpr('hello', ['naw'])
        >>> expr.string
        'naw'
        >>> expr.string = 'huehue'
        >>> expr.string
        'huehue'
        >>> type(expr.string)
        <class 'TexSoup.data.TexText'>
        >>> str(expr)
        "TexExpr('hello', ['huehue'])"
        >>> expr.string = 35  #doctest:+ELLIPSIS
        Traceback (most recent call last):
            ...
        TypeError: ...
        """
        return TexText(''.join(map(str, self._contents)))

    @string.setter
    def string(self, s):
        if not isinstance(s, str):
            raise TypeError(
                '.string value "%s" must be a string or TexText. To set '
                'non-string content, use .contents' % s)
        self.contents = [TexText(s)]

    ##################
    # PUBLIC METHODS #
    ##################

    def append(self, *exprs):
        """Add contents to the expression.

        :param Union[TexExpr,str] exprs: List of contents to add

        >>> expr = TexExpr('textbf', ('hello',))
        >>> expr
        TexExpr('textbf', ['hello'])
        >>> expr.append('world')
        >>> expr
        TexExpr('textbf', ['hello', 'world'])
        """
        self._assert_supports_contents()
        self._contents.extend(exprs)

    def insert(self, i, *exprs):
        """Insert content at specified position into expression.

        :param int i: Position to add content to
        :param Union[TexExpr,str] exprs: List of contents to add

        >>> expr = TexExpr('textbf', ('hello',))
        >>> expr
        TexExpr('textbf', ['hello'])
        >>> expr.insert(0, 'world')
        >>> expr
        TexExpr('textbf', ['world', 'hello'])
        >>> expr.insert(0, TexText('asdf'))
        >>> expr
        TexExpr('textbf', ['asdf', 'world', 'hello'])
        """
        self._assert_supports_contents()
        for j, expr in enumerate(exprs):
            if isinstance(expr, TexExpr):
                expr.parent = self
            self._contents.insert(i + j, expr)

    def remove(self, expr):
        """Remove a provided expression from its list of contents.

        :param Union[TexExpr,str] expr: Content to add
        :return: index of the expression removed
        :rtype: int

        >>> expr = TexExpr('textbf', ('hello',))
        >>> expr.remove('hello')
        0
        >>> expr
        TexExpr('textbf', [])
        """
        self._assert_supports_contents()
        index = self._contents.index(expr)
        self._contents.remove(expr)
        return index

    def _supports_contents(self):
        return True

    def _assert_supports_contents(self):
        pass


class TexEnv(TexExpr):
    r"""Abstraction for a LaTeX command, with starting and ending markers.
    Contains three attributes:

    1. a human-readable environment name,
    2. the environment delimiters
    3. the environment's contents.

    >>> t = TexEnv('displaymath', r'\[', r'\]',
    ...     ['\\mathcal{M} \\circ \\mathcal{A}'])
    >>> t
    TexEnv('displaymath', ['\\mathcal{M} \\circ \\mathcal{A}'], [])
    >>> print(t)
    \[\mathcal{M} \circ \mathcal{A}\]
    >>> len(list(t.children))
    0
    """

    _begin = None
    _end = None

    def __init__(self, name, begin, end, contents=(), args=(),
                 preserve_whitespace=False, position=-1):
        r"""Initialization for Tex environment.

        :param str name: name of environment
        :param str begin: string denoting start of environment
        :param str end: string denoting end of environment
        :param iterable contents: list of contents
        :param iterable args: list of Tex Arguments
        :param bool preserve_whitespace: If false, elements containing only
            whitespace will be removed from contents.
        :param int position: position of first character in original source

        >>> env = TexEnv('math', '$', '$', [r'\$'])
        >>> str(env)
        '$\\$$'
        >>> env.begin = '^^'
        >>> env.end = '**'
        >>> str(env)
        '^^\\$**'
        """
        super().__init__(name, contents, args, preserve_whitespace, position)
        self._begin = begin
        self._end = end

    @property
    def begin(self):
        return self._begin

    @begin.setter
    def begin(self, begin):
        self._begin = begin

    @property
    def end(self):
        return self._end

    @end.setter
    def end(self, end):
        self._end = end

    def __match__(self, name=None, attrs=()):
        """Check if given attributes match environment."""
        if name in (self.name, self.begin +
                    str(self.args), self.begin, self.end):
            return True
        return super().__match__(name, attrs)

    def __str__(self):
        contents = ''.join(map(str, self._contents))
        if self.name == '[tex]':
            return contents
        else:
            return '%s%s%s' % (
                self.begin + str(self.args), contents, self.end)

    def __repr__(self):
        if self.name == '[tex]':
            return str(self._contents)
        if not self.args and not self._contents:
            return "%s('%s')" % (self.__class__.__name__, self.name)
        return "%s('%s', %s, %s)" % (
            self.__class__.__name__, self.name, repr(self._contents),
            repr(self.args))


class TexNamedEnv(TexEnv):
    r"""Abstraction for a LaTeX command, denoted by ``\begin{env}`` and
    ``\end{env}``. Contains three attributes:

    1. the environment name itself,
    2. the environment arguments, whether optional or required, and
    3. the environment's contents.

    **Warning**: Note that *setting* TexNamedEnv.begin or TexNamedEnv.end
    has no effect. The begin and end tokens are always constructed from
    TexNamedEnv.name.

    >>> t = TexNamedEnv('tabular', ['\n0 & 0 & * \\\\\n1 & 1 & * \\\\\n'],
    ...     [BraceGroup('c | c c')])
    >>> t
    TexNamedEnv('tabular', ['\n0 & 0 & * \\\\\n1 & 1 & * \\\\\n'], [BraceGroup('c | c c')])
    >>> print(t)
    \begin{tabular}{c | c c}
    0 & 0 & * \\
    1 & 1 & * \\
    \end{tabular}
    >>> len(list(t.children))
    0
    >>> t = TexNamedEnv('equation', [r'5\sum_{i=0}^n i^2'])
    >>> str(t)
    '\\begin{equation}5\\sum_{i=0}^n i^2\\end{equation}'
    >>> t.name = 'eqn'
    >>> str(t)
    '\\begin{eqn}5\\sum_{i=0}^n i^2\\end{eqn}'
    """

    def __init__(self, name, contents=(), args=(), preserve_whitespace=False,
                 position=-1):
        """Initialization for Tex environment.

        :param str name: name of environment
        :param iterable contents: list of contents
        :param iterable args: list of Tex Arguments
        :param bool preserve_whitespace: If false, elements containing only
            whitespace will be removed from contents.
        :param int position: position of first character in original source
        """
        super().__init__(name, r"\begin{%s}" % name, r"\end{%s}" % name,
                         contents, args, preserve_whitespace, position=position)

    @property
    def begin(self):
        return r"\begin{%s}" % self.name

    @property
    def end(self):
        return r"\end{%s}" % self.name


class TexUnNamedEnv(TexEnv):

    name = None
    begin = None
    end = None

    def __init__(self, contents=(), args=(), preserve_whitespace=False,
                 position=-1):
        """Initialization for Tex environment.

        :param iterable contents: list of contents
        :param iterable args: list of Tex Arguments
        :param bool preserve_whitespace: If false, elements containing only
            whitespace will be removed from contents.
        :param int position: position of first character in original source
        """
        assert self.name, 'Name must be non-falsey'
        assert self.begin and self.end, 'Delimiters must be non-falsey'
        super().__init__(self.name, self.begin, self.end,
                         contents, args, preserve_whitespace, position=position)


class TexDisplayMathModeEnv(TexUnNamedEnv):

    name = '$$'
    begin = '$$'
    end = '$$'
    token_begin = TC.DisplayMathSwitch
    token_end = TC.DisplayMathSwitch


class TexMathModeEnv(TexUnNamedEnv):

    name = '$'
    begin = '$'
    end = '$'
    token_begin = TC.MathSwitch
    token_end = TC.MathSwitch


class TexDisplayMathEnv(TexUnNamedEnv):

    name = 'displaymath'
    begin = r'\['
    end = r'\]'
    token_begin = TC.DisplayMathGroupBegin
    token_end = TC.DisplayMathGroupEnd


class TexMathEnv(TexUnNamedEnv):

    name = 'math'
    begin = r'\('
    end = r'\)'
    token_begin = TC.MathGroupBegin
    token_end = TC.MathGroupEnd


class TexCmd(TexExpr):
    r"""Abstraction for a LaTeX command. Contains two attributes:

    1. the command name itself and
    2. the command arguments, whether optional or required.

    >>> textit = TexCmd('textit', args=[BraceGroup('slant')])
    >>> t = TexCmd('textbf', args=[BraceGroup('big ', textit, '.')])
    >>> t
    TexCmd('textbf', [BraceGroup('big ', TexCmd('textit', [BraceGroup('slant')]), '.')])
    >>> print(t)
    \textbf{big \textit{slant}.}
    >>> children = list(map(str, t.children))
    >>> len(children)
    1
    >>> print(children[0])
    \textit{slant}
    """

    def __str__(self):
        if self._contents:
            return '\\%s%s%s' % (self.name, self.args, ''.join(
                [str(e) for e in self._contents]))
        return '\\%s%s' % (self.name, self.args)

    def __repr__(self):
        if not self.args:
            return "TexCmd('%s')" % self.name
        return "TexCmd('%s', %s)" % (self.name, repr(self.args))

    def _supports_contents(self):
        return self.name == 'item'

    def _assert_supports_contents(self):
        if not self._supports_contents():
            raise TypeError(
                'Command "{}" has no children. `add_contents` is only valid '
                'for: 1. environments like `itemize` and 2. `\\item`. '
                'Alternatively, you can add, edit, or delete arguments by '
                'modifying `.args`, which behaves like a list.'
                .format(self.name))


class TexText(TexExpr, str):
    r"""Abstraction for LaTeX text.

    Representing regular text objects in the parsed tree allows users to
    search and modify text objects as any other expression allows.

    >>> obj = TexNode(TexText('asdf gg'))
    >>> 'asdf' in obj
    True
    >>> 'err' in obj
    False
    >>> TexText('df ').strip()
    'df'
    """

    _has_custom_contain = True

    def __init__(self, text, position=-1):
        """Initialize text as tex expresssion.

        :param str text: Text content
        :param int position: position of first character in original source
        """
        super().__init__('text', [text], position=position)
        self._text = text

    def __contains__(self, other):
        """
        >>> obj = TexText(Token('asdf'))
        >>> 'a' in obj
        True
        >>> 'b' in obj
        False
        """
        return other in self._text

    def __eq__(self, other):
        """
        >>> TexText('asdf') == 'asdf'
        True
        >>> TexText('asdf') == TexText('asdf')
        True
        >>> TexText('asfd') == 'sdddsss'
        False
        """
        if isinstance(other, TexText):
            return self._text == other._text
        if isinstance(other, str):
            return self._text == other
        return False

    def __str__(self):
        """
        >>> TexText('asdf')
        'asdf'
        """
        return str(self._text)

    def __repr__(self):
        """
        >>> TexText('asdf')
        'asdf'
        """
        return repr(self._text)


#############
# Arguments #
#############


class TexGroup(TexUnNamedEnv):
    """Abstraction for a LaTeX environment with single-character delimiters.

    Used primarily to identify and associate arguments with commands.
    """

    def __init__(self, *contents, preserve_whitespace=False, position=-1):
        """Initialize argument using list of expressions.

        :param Union[str,TexCmd,TexEnv] exprs: Tex expressions contained in the
            argument. Can be other commands or environments, or even strings.
        :param int position: position of first character in original source
        """
        super().__init__(contents, preserve_whitespace=preserve_whitespace,
                         position=position)

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__,
                           ', '.join(map(repr, self._contents)))

    @classmethod
    def parse(cls, s):
        """Parse a string or list and return an Argument object.

        Naive implementation, does not parse expressions in provided string.

        :param Union[str,iterable] s: Either a string or a list, where the
            first and last elements are valid argument delimiters.

        >>> TexGroup.parse('[arg0]')
        BracketGroup('arg0')
        """
        assert isinstance(s, str)
        for arg in arg_type:
            if s.startswith(arg.begin) and s.endswith(arg.end):
                return arg(s[len(arg.begin):-len(arg.end)])
        raise TypeError('Malformed argument: %s. Must be an TexGroup or a string in'
                        ' either brackets or curly braces.' % s)


class BracketGroup(TexGroup):
    """Optional argument, denoted as ``[arg]``"""

    begin = '['
    end = ']'
    name = 'BracketGroup'
    token_begin = TC.BracketBegin
    token_end = TC.BracketEnd


class BraceGroup(TexGroup):
    """Required argument, denoted as ``{arg}``."""

    begin = '{'
    end = '}'
    name = 'BraceGroup'
    token_begin = TC.GroupBegin
    token_end = TC.GroupEnd


arg_type = (BracketGroup, BraceGroup)


class TexArgs(list):
    r"""List of arguments for a TeX expression. Supports all standard list ops.

    Additional support for conversion from and to unparsed argument strings.

    >>> arguments = TexArgs(['\n', BraceGroup('arg0'), '[arg1]', '{arg2}'])
    >>> arguments
    [BraceGroup('arg0'), BracketGroup('arg1'), BraceGroup('arg2')]
    >>> arguments.all
    ['\n', BraceGroup('arg0'), BracketGroup('arg1'), BraceGroup('arg2')]
    >>> arguments[2]
    BraceGroup('arg2')
    >>> len(arguments)
    3
    >>> arguments[:2]
    [BraceGroup('arg0'), BracketGroup('arg1')]
    >>> isinstance(arguments[:2], TexArgs)
    True
    """

    def __init__(self, args=[]):
        """List of arguments for a command.

        :param list args: List of parsed or unparsed arguments
        """
        super().__init__()
        self.all = []
        self.extend(args)

    def __coerce(self, arg):
        if isinstance(arg, str) and not arg.isspace():
            arg = TexGroup.parse(arg)
        return arg

    def append(self, arg):
        """Append whitespace, an unparsed argument string, or an argument
        object.

        :param TexGroup arg: argument to add to the end of the list

        >>> arguments = TexArgs([BraceGroup('arg0'), '[arg1]', '{arg2}'])
        >>> arguments.append('[arg3]')
        >>> arguments[3]
        BracketGroup('arg3')
        >>> arguments.append(BraceGroup('arg4'))
        >>> arguments[4]
        BraceGroup('arg4')
        >>> len(arguments)
        5
        >>> arguments.append('\\n')
        >>> len(arguments)
        5
        >>> len(arguments.all)
        6
        """
        self.insert(len(self), arg)

    def extend(self, args):
        """Extend mixture of unparsed argument strings, arguments objects, and
        whitespace.

        :param List[TexGroup] args: Arguments to add to end of the list

        >>> arguments = TexArgs([BraceGroup('arg0'), '[arg1]', '{arg2}'])
        >>> arguments.extend(['[arg3]', BraceGroup('arg4'), '\\t'])
        >>> len(arguments)
        5
        >>> arguments[4]
        BraceGroup('arg4')
        """
        for arg in args:
            self.append(arg)

    def insert(self, i, arg):
        r"""Insert whitespace, an unparsed argument string, or an argument
        object.

        :param int i: Index to insert argument into
        :param TexGroup arg: Argument to insert

        >>> arguments = TexArgs(['\n', BraceGroup('arg0'), '[arg2]'])
        >>> arguments.insert(1, '[arg1]')
        >>> len(arguments)
        3
        >>> arguments
        [BraceGroup('arg0'), BracketGroup('arg1'), BracketGroup('arg2')]
        >>> arguments.all
        ['\n', BraceGroup('arg0'), BracketGroup('arg1'), BracketGroup('arg2')]
        >>> arguments.insert(10, '[arg3]')
        >>> arguments[3]
        BracketGroup('arg3')
        """
        arg = self.__coerce(arg)

        if isinstance(arg, (TexGroup, TexCmd)):
            super().insert(i, arg)

        if len(self) <= 1:
            self.all.append(arg)
        else:
            if i > len(self):
                i = len(self) - 1

            before = self[i - 1]
            index_before = self.all.index(before)
            self.all.insert(index_before + 1, arg)

    def remove(self, item):
        """Remove either an unparsed argument string or an argument object.

        :param Union[str,TexGroup] item: Item to remove

        >>> arguments = TexArgs([BraceGroup('arg0'), '[arg2]', '{arg3}'])
        >>> arguments.remove('{arg0}')
        >>> len(arguments)
        2
        >>> arguments[0]
        BracketGroup('arg2')
        >>> arguments.remove(arguments[0])
        >>> arguments[0]
        BraceGroup('arg3')
        >>> arguments.remove(BraceGroup('arg3'))
        >>> len(arguments)
        0
        >>> arguments = TexArgs([
        ...     BraceGroup(TexCmd('color')),
        ...     BraceGroup(TexCmd('color', [BraceGroup('blue')]))
        ... ])
        >>> arguments.remove(arguments[0])
        >>> len(arguments)
        1
        >>> arguments.remove(arguments[0])
        >>> len(arguments)
        0
        """
        item = self.__coerce(item)
        self.all.remove(item)
        super().remove(item)

    def pop(self, i):
        """Pop argument object at provided index.

        :param int i: Index to pop from the list

        >>> arguments = TexArgs([BraceGroup('arg0'), '[arg2]', '{arg3}'])
        >>> arguments.pop(1)
        BracketGroup('arg2')
        >>> len(arguments)
        2
        >>> arguments[0]
        BraceGroup('arg0')
        """
        item = super().pop(i)
        j = self.all.index(item)
        return self.all.pop(j)

    def reverse(self):
        r"""Reverse both the list and the proxy `.all`.

        >>> args = TexArgs(['\n', BraceGroup('arg1'), BracketGroup('arg2')])
        >>> args.reverse()
        >>> args.all
        [BracketGroup('arg2'), BraceGroup('arg1'), '\n']
        >>> args
        [BracketGroup('arg2'), BraceGroup('arg1')]
        """
        super().reverse()
        self.all.reverse()

    def clear(self):
        r"""Clear both the list and the proxy `.all`.

        >>> args = TexArgs(['\n', BraceGroup('arg1'), BracketGroup('arg2')])
        >>> args.clear()
        >>> len(args) == len(args.all) == 0
        True
        """
        super().clear()
        self.all.clear()

    def __getitem__(self, key):
        """Standard list slicing.

        Returns TexArgs object for subset of list and returns an TexGroup object
        for single items.

        >>> arguments = TexArgs([BraceGroup('arg0'), '[arg1]', '{arg2}'])
        >>> arguments[2]
        BraceGroup('arg2')
        >>> arguments[:2]
        [BraceGroup('arg0'), BracketGroup('arg1')]
        """
        value = super().__getitem__(key)
        if isinstance(value, list):
            return TexArgs(value)
        return value

    def __contains__(self, item):
        """Checks for membership. Allows string comparisons to args.

        >>> arguments = TexArgs(['{arg0}', '[arg1]'])
        >>> 'arg0' in arguments
        True
        >>> BracketGroup('arg0') in arguments
        False
        >>> BraceGroup('arg0') in arguments
        True
        >>> 'arg3' in arguments
        False
        """
        if isinstance(item, str):
            return any([item == arg.string for arg in self])
        return super().__contains__(item)

    def __str__(self):
        """Stringifies a list of arguments.

        >>> str(TexArgs(['{a}', '[b]', '{c}']))
        '{a}[b]{c}'
        """
        return ''.join(map(str, self))

    def __repr__(self):
        """Makes list of arguments command-line friendly.

        >>> TexArgs(['{a}', '[b]', '{c}'])
        [BraceGroup('a'), BracketGroup('b'), BraceGroup('c')]
        """
        return '[%s]' % ', '.join(map(repr, self))



import bisect
import functools

from enum import IntEnum as IntEnumBase


##########
# Tokens #
##########


def IntEnum(name, keys, start=1):
    """Explicitly define key-value pairs. For Python3.4 compatibility"""
    return IntEnumBase(name,
                       [(key, index) for index, key in enumerate(keys, start=start)])


CC = IntEnum('CategoryCodes', (
    'Escape',
    'GroupBegin',
    'GroupEnd',
    'MathSwitch',
    'Alignment',
    'EndOfLine',
    'Macro',
    'Superscript',
    'Subscript',
    'Ignored',
    'Spacer',
    'Letter',
    'Other',
    'Active',
    'Comment',
    'Invalid',

    # custom
    'MathGroupBegin',
    'MathGroupEnd',
    'BracketBegin',
    'BracketEnd',
    'ParenBegin',
    'ParenEnd'
))


# Only includes items that cannot cause failures
TC = IntEnum('TokenCode', (
    'Escape',
    'GroupBegin',
    'GroupEnd',
    'Comment',
    'MergedSpacer',  # whitespace allowed between <command name> and arguments
    'EscapedComment',
    'MathSwitch',
    'DisplayMathSwitch',
    'MathGroupBegin',
    'MathGroupEnd',
    'DisplayMathGroupBegin',
    'DisplayMathGroupEnd',
    'LineBreak',
    'CommandName',
    'Text',
    'BracketBegin',
    'BracketEnd',
    'ParenBegin',
    'ParenEnd',

    # temporary (Replace with macros support)
    'PunctuationCommandName',
    'SizeCommand',
    'Spacer'
), start=max(CC))


class Token(str):
    """Enhanced string object with knowledge of global position."""

    # noinspection PyArgumentList
    def __new__(cls, text='', position=None, category=None):
        """Initializer for pseudo-string object.

        :param text: The original string
        :param position: Position in the original buffer
        :param category: Category of token
        """
        self = str.__new__(cls, text)
        if isinstance(text, Token):
            self.text = text.text
            self.position = text.position
            self.category = category or text.category
        else:
            self.text = text
            self.position = position
            self.category = category
        return self

    def __repr__(self):
        return repr(self.text)

    def __str__(self):
        return str(self.text)

    def __getattr__(self, name):
        return getattr(self.text, name)

    def __eq__(self, other):
        """
        >>> Token('asdf', 0) == Token('asdf', 2)
        True
        >>> Token('asdf', 0) == Token('asd', 0)
        False
        """
        if isinstance(other, Token):
            return self.text == other.text
        else:
            return self.text == other

    def __hash__(self):
        """
        >>> hash(Token('asf')) == hash('asf')
        True
        """
        return hash(self.text)

    def __add__(self, other):
        """Implements addition in the form of TextWithPosition(...) + (obj).

        >>> t1 = Token('as', 0) + Token('df', 1)
        >>> str(t1)
        'asdf'
        >>> t1.position
        0
        >>> t2 = Token('as', 1) + 'df'
        >>> str(t2)
        'asdf'
        >>> t3 = Token(t2)
        >>> t3.position
        1
        """
        if isinstance(other, Token):
            return Token(self.text + other.text, self.position, self.category)
        else:
            return Token(self.text + other, self.position, self.category)

    def __radd__(self, other):
        """Implements addition in the form of (obj) + TextWithPosition(...).

        Note that if the first element is Token,
        Token(...).__add__(...) will be used. As a result, we
        can assume WLOG that `other` is a type other than Token.

        >>> t1 = Token('as', 2) + Token('dfg', 2)
        >>> str(t1)
        'asdfg'
        >>> t1.position
        2
        >>> t2 = 'as' + Token('dfg', 2)
        >>> str(t2)
        'asdfg'
        >>> t2.position
        0
        """
        return Token(
            other + self.text, self.position - len(other), self.category)

    def __iadd__(self, other):
        """Implements addition in the form of TextWithPosition(...) += ...

        >>> t1 = Token('as', 0)
        >>> t1 += 'df'
        >>> str(t1)
        'asdf'
        >>> t1.position
        0
        """
        if isinstance(other, Token):
            new = Token(self.text + other.text, self.position, self.category)
        else:
            new = Token(self.text + other, self.position, self.category)
        return new

    @classmethod
    def join(cls, tokens, glue=''):
        if len(tokens) > 0:
            return Token(
                glue.join(t.text for t in tokens),
                tokens[0].position,
                tokens[0].category)
        else:
            return Token.Empty

    def __bool__(self):
        return bool(self.text)

    def __contains__(self, item):
        """
        >>> 'rg' in Token('corgi', 0)
        True
        >>> 'reg' in Token('corgi', 0)
        False
        >>> Token('rg', 0) in Token('corgi', 0)
        True
        """
        if isinstance(item, Token):
            return item.text in self.text
        return item in self.text

    def __iter__(self):
        """
        >>> list(Token('asdf', 0))
        ['a', 's', 'd', 'f']
        """
        return iter(self.__iter())

    def __iter(self):
        for i, c in enumerate(self.text):
            yield Token(c, self.position + i, self.category)

    def __getitem__(self, i):
        """Access characters in object just as with strings.

        >>> t1 = Token('asdf', 2)
        >>> t1[0]
        'a'
        >>> t1[-1]
        'f'
        >>> t1[:]
        'asdf'
        """
        if isinstance(i, int):
            start = i
        else:
            start = i.start
        if start is None:
            start = 0
        if start < 0:
            start = len(self.text) + start
        return Token(self.text[i], self.position + start, self.category)

    def strip(self, *args, **kwargs):
        stripped = self.text.strip(*args, **kwargs)
        offset = self.text.find(stripped)
        return Token(stripped, self.position + offset, self.category)

    def lstrip(self, *args, **kwargs):
        """Strip leading whitespace for text.

        >>> t = Token('  asdf  ', 2)
        >>> t.lstrip()
        'asdf  '
        """
        stripped = self.text.lstrip(*args, **kwargs)
        offset = self.text.find(stripped)
        return Token(stripped, self.position + offset, self.category)

    def rstrip(self, *args, **kwargs):
        """Strip trailing whitespace for text.

        >>> t = Token('  asdf  ', 2)
        >>> t.rstrip()
        '  asdf'
        """
        stripped = self.text.rstrip(*args, **kwargs)
        offset = self.text.find(stripped)
        return Token(stripped, self.position + offset, self.category)


Token.Empty = Token('', position=0)


# TODO: Rename to Buffer (formerly MixedBuffer) and StringBuffer
# but needs test refactoring to change defaults
class Buffer:
    """Converts string or iterable into a navigable iterator of strings.

    >>> b1 = Buffer("012345")
    >>> next(b1)
    '0'
    >>> b1.forward()
    '1'
    >>> b1.endswith('1')
    True
    >>> b1.backward(2)
    '01'
    >>> b1.peek()
    '0'
    >>> b1.peek(2)
    '2'
    >>> b1.peek((0, 2))
    '01'
    >>> b1.startswith('01')
    True
    >>> b1[2:4]
    '23'
    >>> Buffer('asdf')[:10]
    'asdf'
    >>> def gen():
    ...     for i in range(10):
    ...         yield i
    >>> list(gen())
    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    >>> list(Buffer(gen()))
    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    """

    def __init__(self, iterator, join=Token.join, empty=lambda: '',
                 init=lambda content, index: Token(content, index)):
        """Initialization for Buffer.

        :param iterator: iterator or iterable
        :param func join: function to join multiple buffer elements
        """
        assert hasattr(iterator, '__iter__'), 'Must be an iterable.'
        self.__iterator = iter(iterator)
        self.__queue = []
        self.__i = 0
        self.__join = join
        self.__init = init
        self.__empty = empty

    # noinspection PyPep8Naming
    def hasNext(self, n=1):
        """Returns whether or not there is another element."""
        return bool(self.peek(n - 1))

    def startswith(self, s):
        """Check if iterator starts with s, beginning from the current
        position."""
        return self.peek((0, len(s))).startswith(s)

    def endswith(self, s):
        """Check if iterator ends with s, ending at current position."""
        return self.peek((-len(s), 0)).endswith(s)

    def forward(self, j=1):
        """Move forward by j steps.

        >>> b = Buffer('abcdef')
        >>> b.forward(3)
        'abc'
        >>> b.forward(-2)
        'bc'
        """
        if j < 0:
            return self.backward(-j)
        self.__i += j
        return self[self.__i - j:self.__i]

    def num_forward_until(self, condition):
        """Forward until one of the provided matches is found.

        :param condition: set of valid strings
        """
        i, c = 0, ''
        while self.hasNext() and not condition(self.peek()):
            c += self.forward(1)
            i += 1
        assert self.backward(i) == c
        return i

    def forward_until(self, condition, peek=True):
        """Forward until one of the provided matches is found.

        The returned string contains all characters found before the condition
        was met. In other words, the condition will be true for the remainder
        of the buffer.

        :param Callable condition: lambda condition for the token to stop at

        >>> buf = Buffer(map(str, range(9)))
        >>> _ = buf.forward_until(lambda x: int(x) > 3)
        >>> c = buf.forward_until(lambda x: int(x) > 6)
        >>> c
        '456'
        >>> c.position
        4
        """
        c = self.__init(self.__empty(), self.peek().position)
        while self.hasNext() and not condition(self.peek() if peek else self):
            c += self.forward(1)
        return c

    def backward(self, j=1):
        """Move backward by j steps.

        >>> b = Buffer('abcdef')
        >>> b.backward(-3)
        'abc'
        >>> b.backward(2)
        'bc'
        """
        if j < 0:
            return self.forward(-j)
        assert self.__i - j >= 0, 'Cannot move more than %d back' % self.__i
        self.__i -= j
        return self[self.__i:self.__i + j]

    def peek(self, j=0):
        """Peek at the next value(s), without advancing the Buffer.

        Return None if index is out of range.
        """
        try:
            if isinstance(j, int):
                return self[self.__i + j]
            return self[self.__i + j[0]:self.__i + j[1]]
        except IndexError:
            return None

    def __next__(self):
        """Implements next."""
        while self.__i >= len(self.__queue):
            self.__queue.append(self.__init(
                next(self.__iterator), self.__i))
        self.__i += 1
        return self.__queue[self.__i - 1]

    def __getitem__(self, i):
        """Supports indexing list.

        >>> b = Buffer('asdf')
        >>> b[5]
        Traceback (most recent call last):
            ...
        IndexError: list index out of range
        >>> b[0]
        'a'
        >>> b[1:3]
        'sd'
        >>> b[1:]
        'sdf'
        >>> b[:3]
        'asd'
        >>> b[:]
        'asdf'
        """
        if isinstance(i, int):
            old, j = self.__i, i
        else:
            old, j = self.__i, i.stop

        while j is None or self.__i <= j:
            try:
                next(self)
            except StopIteration:
                break
        self.__i = old
        if isinstance(i, int):
            return self.__queue[i]
        return self.__join(self.__queue[i])

    def __iter__(self):
        return self

    @property
    def position(self):
        return self.__i


class CharToLineOffset(object):
    """Utility to convert absolute position in the source file to
    line_no:char_no_in_line. This can be very useful if we want to parse LaTeX
    and navigate to some elements in the generated DVI/PDF via SyncTeX.

    >>> clo = CharToLineOffset('''hello
    ... world
    ... I scream for ice cream!''')
    >>> clo(3)
    (0, 3)
    >>> clo(6)
    (1, 0)
    >>> clo(12)
    (2, 0)
    """

    def __init__(self, src):
        self.line_break_positions = [i for i, c in enumerate(src) if c == '\n']
        self.src_len = len(src)

    def __call__(self, char_pos):
        line_no = bisect.bisect(self.line_break_positions, char_pos)
        if line_no == 0:
            char_no = char_pos
        elif line_no == len(self.line_break_positions):
            line_start = self.line_break_positions[-1]
            char_no = min(char_pos - line_start - 1, self.src_len - line_start)
        else:
            char_no = char_pos - self.line_break_positions[line_no - 1] - 1
        return line_no, char_no


class MixedBuffer(Buffer):

    def __init__(self, iterator):
        """Initialization for Buffer, accepting types beyond strings.

        :param iterator: iterator or iterable
        :param func join: function to join multiple buffer elements

        >>> buf = MixedBuffer([324, 'adsf', lambda x: x])
        >>> buf.peek()
        324
        """
        super().__init__(iterator,
                         join=lambda x: x, empty=lambda x: [],
                         init=lambda content, index: content)


##############
# Decorators #
##############


def to_buffer(convert_in=True, convert_out=True, Buffer=Buffer):
    """Decorator converting all strings and iterators/iterables into
    Buffers.

    :param bool convert_in: Convert inputs where applicable to Buffers
    :param bool convert_out: Convert output to a Buffer
    :param type Buffer: Type of Buffer to convert into
    """
    def decorator(f):
        @functools.wraps(f)
        def wrap(*args, **kwargs):
            iterator = args[0]
            if convert_in:
                iterator = kwargs.get('iterator', iterator)
                if not isinstance(iterator, Buffer):
                    iterator = Buffer(iterator)
            output = f(iterator, *args[1:], **kwargs)
            if convert_out:
                return Buffer(output)
            return output
        return wrap
    return decorator


def to_list(f):
    """Converts generator or iterable output to list

    >>> class A:
    ...     @property
    ...     @to_list
    ...     def a(self):
    ...         for i in range(3):
    ...             yield i
    >>> A().a
    [0, 1, 2]
    """
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        return list(f(*args, **kwargs))
    return wrapper



from .convertNoyer import run
import argparse

argparser = argparse.ArgumentParser(
    description='Convertit un fichier .tex en un fichier .md')

argparser.add_argument('input', metavar='input', type=str,
                          help='le fichier .tex à convertir')

def main():
    args = argparser.parse_args()
    run(filename=args.input)


#!/usr/bin/env python3
'''
 Name : Elowan
 Creation : 24-06-2023 20:53:52
 Last modified : 28-06-2023 15:00:51
'''
from .TexSoup import TexSoup, utils, data
import re
import os

from .CONST import disponible_environments, skip_envs
from .latex2png import pnglatex

# Soup = l'intérieur d'une balise

def frameAnalyse(soup):
    """
    Analyse le contenu d'un frame et le transforme en markdown

    (Prend en compte les frames fragiles et les titres pas dans des 
    frametitle)

    :param soup: le contenu du frame
    :type soup: TexSoup.TexNode
    :return: le contenu du frame en markdown
    """
    # Si la frame est sans paramètres
    if len(soup.args) == 0:
        return transformSoupToMarkdown(list(soup.contents))
    
    # Si la frame est fragile
    if soup.args[0] == '[fragile]':
        string = ''

        # Si la frame a un titre
        if len(soup.args) >= 2:
            string += "\n#### " + removeWeirdSpacing(str(soup.args[1])) \
                    .replace('{', '').replace('}', '') + "\n\n"
            return string + transformSoupToMarkdown(list(soup.contents)[2:])
        
        else:
            return "\n" + transformSoupToMarkdown(list(soup.contents)[1:])
    else:
        # Si la frame a un titre mais pas de mention fragile
        string = "\n#### " + removeWeirdSpacing(str(soup.args[0])).replace('{', '').replace('}', '') + "\n\n"
        return string + transformSoupToMarkdown(list(soup.contents)[1:])
    
def getEnvironment(string):
    """
    Récupère le contenu d'un environnement dans un string

    Trouve le premier '{' et renvoie tout ce qu'il y a dedans 

    :param string: le string à analyser
    :type string: str
    :return: le contenu de l'environnement
    """
    count = None
    ignoreLine = False
    for i in range(len(string)):
        if ignoreLine:
            continue

        if string[i] == '%':
            ignoreLine = True
            continue

        if string[i] == '{':
            if count == None: count = 0
            count += 1
        elif string[i] == '}':
            count -= 1
        if count == 0:
            return string[string.find('{')+1:i]

def getWeirdEnvironment(string):
    """
    Récupère le contenu d'une commande de la forme "\env ... {}" 
    dans un string avec ... les options de la commande

    :param string: le string à analyser
    :type string: str
    :return: (
        la commande, 
        le contenu de l'environnement, 
        la commande complète
    )
    """
    command = ""

    # Récupère la commande latex
    for i in range(len(string)):
        if string[i] == '\\':
            command += string[i:string.find('{')]
            break

    # Récupère le contenu de l'environnement
    envContent = getEnvironment(string[len(command):])
    
    return string[string.find('\\'):len(command)+len(envContent)+3]

def renderImage(name, string):
    """
    Rend une image latex en png

    :param name: le nom de l'image
    :type name: str
    :param string: le contenu de l'image
    :type string: str
    :return: le nom de l'image
    :rtype: str
    """
    print("/!\ Commande impossible à parser : {}".format(
                removeWeirdSpacing(string)))
    print("Conversion en image...\n")
    
    imageNumber = len([name for name in os.listdir(DIR) if os.path.isfile(os.path.join(DIR, name))])

    # Convertit le latex en image
    try:
        filename = "{}_{}.png".format(name, imageNumber)
        pnglatex(string, DIR + "{}".format(filename), packages=packages)

    except ValueError:
        print("/!\ Erreur lors de la conversion en image\n")
        return "\n!!!danger \"Erreur de convertion : {}\"\n".format(
                removeWeirdSpacing(filename))
    

    else:
        print("Conversion en image réussie\n")
        return "\n\n![{}]({})\n\n".format(name, DIR + "{}".format(filename))

# Fonctions de transformation de balise en markdown
functions_for_envs = {
    # MetaBalises
    "input": lambda x, indent: "\n\n!!!danger \"Include\"\n", # TODO: include the file
    "frame": lambda x, indent: frameAnalyse(x),
    "title": lambda x, indent: "# " + removeWeirdSpacing(str(x.string)) + "\n\n",
    "section": lambda x, indent: "\n## " + removeWeirdSpacing(str(x.string)) + "\n\n",
    "subsection": lambda x, indent: "\n### " + removeWeirdSpacing(str(x.string)) + "\n\n",
    "frametitle": lambda x, indent: "\n\n#### " + removeWeirdSpacing(str(x.string)) + "\n\n",
    "framesubtitle": lambda x, indent: "\n##### " + removeWeirdSpacing(str(x.string)) + "\n\n",

    
    # Blocs à part entière
    "label": lambda x, indent: "", # On dégage les labels
    "tiny": lambda x, indent: str(transformSoupToMarkdown(list(x.contents))),
    "small": lambda x, indent: str(transformSoupToMarkdown(list(x.contents))),
    "ref": lambda x, indent: "[" + removeWeirdSpacing(str(x.string)) + "](#"
                            + str(x.string) + ")",
    "tabular": lambda x, indent: renderImage(str(x.name), str(x)),

    "figure": lambda x, indent: "\n" + renderImage(x.name, str(x)),
    
    "itemize": lambda x, indent: "\n\n" + "\n".join([indent + "- " + str(
                                    transformSoupToMarkdown(list(elmt))
                                    ).rstrip().lstrip() 
                                    for elmt in list(x.children)]) + "\n\n",

    "enumerate": lambda x, indent: "\n\n" + "\n".join([indent + "1. " + str(
                                    transformSoupToMarkdown(list(elmt), indent)
                                    ).rstrip().lstrip()
                                    for elmt in list(x.children)]) + "\n\n",

    "lstlisting": lambda x, indent: "\n```{} linenums=\"1\"\n".format(str(x.contents[0])
                                    .replace("style=", "")) + 
                                    str(x.contents[1]).replace('    ', ' ') + 
                                    "\n```\n",

    "columns": lambda x, indent: str(transformSoupToMarkdown(list(x.contents[2:]), indent)),
    "column": lambda x, indent: str(transformSoupToMarkdown(list(x.contents[1:]), indent)),

    "defn": lambda x, indent: '\n\n!!!quote "Définition"\n' + "     " + 
                                    str(transformSoupToMarkdown(list(x.contents), 
                                    indent + "     ")) + "\n\n",

    "rem": lambda x, indent: '\n\n!!!tip ""\n    **Remarque**\n\n' + "     " +
                                    str(transformSoupToMarkdown(list(x.contents), 
                                    indent + "     ")) + "\n\n",

    "prop": lambda x, indent: '\n\n!!!warning ""\n    **Propriété**\n\n' + "     " +
                                    str(transformSoupToMarkdown(list(x.contents), 
                                    indent + "     ")) + "\n\n",
    "block": lambda x, indent: "\n\n" + str(transformSoupToMarkdown(list(x.contents),
                                    indent)) + "\n\n",
    "cod": lambda x, indent: "\n`" + str(transformSoupToMarkdown(list(x.contents),
                                    indent)) + "`\n",
    

    # Inline elements
    "href": lambda x, indent: "[" + removeWeirdSpacing(str(x.args[1].string)) + 
                                    "](" + str(x.args[0].string) + ")",

    "emph": lambda x, indent: " *" + str(transformSoupToMarkdown(list(x.contents)
                                    )) + "*",
    "enquote": lambda x, indent: "\"" + str(transformSoupToMarkdown(list(x.contents)
                                    )) + "\"",
    "textcolor": lambda x, indent: "<span style='color:{}'>".format(
                                    str(x.args[0].string)
                                    ) + str(transformSoupToMarkdown(
                                        list(x.contents[1:]))
                                    ) + "</span>",

    "displaymath": lambda x, indent: "\n\n$$\n" + " ".join(
                                    [ str(x).rstrip().lstrip() for x in x.contents]
                                    ) + "$$\n\n",
    "BraceGroup": lambda x, indent: str(transformSoupToMarkdown(
                                    list(x.contents), indent)),

    # "alert": lambda x, indent: str(transformSoupToMarkdown(
    #                                 list(TexSoup(getEnvironment(stringSoup[x.position:]), 
    #                                 skip_envs=skip_envs)), indent)),

    "textbf": lambda x, indent: "**" + str(transformSoupToMarkdown(
                                    list(x.contents))
                                    ) + "**",
    
    "gbox": lambda x, indent: " `" + str(transformSoupToMarkdown(
                                    list(x.contents))
                                    ) + "`",
    "textwidth": lambda x, indent: str(transformSoupToMarkdown(
                                    list(x.contents))
                                    ),

    # Abréviations (TODO: à vérifier)
    "node": lambda x, indent: "noeud",
    "ie": lambda x, indent: "ie",
    "ssi": lambda x, indent: "si et seulement si",
    "": lambda x, indent: transformSoupToMarkdown(list(x.contents), indent),
}


def transformSoupToMarkdown(soupList, indent=""):
    """
    Transforme une liste de soup en markdown, si un % est placé au début de la ligne,
    alors la ligne n'est pas gardée. La fonction est récursive.

    :param list: une liste de soup
    :type list: List

    :return: string
    """
    resultString = ''
    toSkip = -1

    # Parcours de la liste de soup
    for i in range(len(soupList)):
        soup = soupList[i]

        if soup.position < toSkip:
            continue

        # Si c'est un commentaire, on passe à la ligne suivante
        if str(soup).startswith('%'):
            continue

        # Si c'est un simple texte ou des maths, on le retourne
        if type(soup) == utils.Token or soup.name == "$":
            resultString += ' ' + removeWeirdSpacing(str(soup))

            if type(soup) != data.TexNode:
                resultString = resultString.rstrip().lstrip()
                
            continue

        # Si on a prévu une fonction pour cette balise, on l'applique
        if soup.name in functions_for_envs:
            resultString += str(functions_for_envs[soup.name](soup, indent))
            continue
        
        # Si la balise n'est pas prise en charge par le script
        # qui se chargera de l'afficher sur le site, on affiche un warning
        if soup.name not in disponible_environments:
            string = getWeirdEnvironment(stringSoup[soup.position-1:])

            # Transforme la balise en image             
            string = renderImage(soup.name, string)

            toSkip = soup.position + len(string) + 1
            
            resultString += string

            # resultString += "\n\n$$\n" + removeWeirdSpacing(str(soup)).rstrip().lstrip() + "\n$$\n\n"
            continue

        # Sinon on applique la fonction sur tous les enfants de la balise
        resultString += transformSoupToMarkdown(list(soup.contents), indent)

        resultString += "\n"
        
    return resultString

def removeWeirdSpacing(string):
    """
    Enlève les espaces/saut à la ligne en trop dans une string

    :param string: la string à modifier
    :type string: string

    :return: string
    """
    # On enlève les sauts à la ligne, les espaces et les % en trop
    reg1 = re.compile(r'\s+')
    reg2 = re.compile(r'[\n\r%]')
    string = re.sub(reg1, ' ', string)
    return re.sub(reg2, '', string)

def includeFiles(soup):
    """
    Include les fichiers .tex dans le fichier principal

    :param soup: le fichier principal
    :type soup: TexSoup

    :return: string
    """
    # Récupère les fichiers à inclure
    inputs = soup.find_all('input')
    for input_soup in inputs:
        # Ouvre le fichier
        try:
            with open(str(input_soup.args[0].string), 'r') as f:
                stringSoup = f.read()

            # Ajoute le contenu du fichier au fichier principal
            input_soup.replace_with(TexSoup(stringSoup, skip_envs=skip_envs))
        
        except FileNotFoundError:
            print("/!\ Impossible d'importer le fichier " + 
                  str(input_soup.args[0].string))
            
    return TexSoup(str(soup), skip_envs=skip_envs)

def getAllImportedPackages(soup):
    """
    Récupère tous les packages importés dans le fichier

    :param soup: le fichier principal
    :type soup: TexSoup

    :return: List
    """
    packages = []
    for package in soup.find_all('usepackage'):
        packages.append(str(package))
    return packages

def removeComments(soup):
    """
    Enlève les commentaires du fichier

    :param soup: le fichier principal
    :type soup: TexSoup

    :return: string
    """
    # Récupère tous les commentaires
    ignoreLine = False
    soupstr = str(soup)
    outstring = ""
    for i in range(len(soupstr)):
        if soupstr[i] == '%':
            ignoreLine = True
            continue

        if soupstr[i] == '\n':
            ignoreLine = False
            
        if not ignoreLine:
            outstring += soupstr[i]

    return TexSoup(outstring)

def removeOldImages():
    """
    Supprime les images précédentes
    """
    if os.path.exists(DIR):
        for file in os.listdir(DIR):
            os.remove(DIR + file)

def run(filename):
    """
    Fonction principale du script, transforme un fichier .tex en markdown
    
    :param filename: le nom du fichier à transformer
    :type filename: string
    """
    global stringSoup
    global packages 

    # Ouverture du fichier
    with open(filename, 'r') as f:
        stringSoup = f.read()
        
    soup = TexSoup(stringSoup, skip_envs=skip_envs)
    soup = includeFiles(soup)
    soup = removeComments(soup)

    stringSoup = str(soup)

    # Supprime les images précédentes
    removeOldImages()
    os.makedirs(DIR, exist_ok=True)

    # Récupère tous les packages importés
    packages = getAllImportedPackages(soup)

    # Récupération de la partie importante du fichier
    document = soup.find('document')
    documentList = list(document.contents)

    # Transformation en markdown
    outputString = ""

    # Ajout du titre, author et date
    outputString += "# " + removeWeirdSpacing(str(soup.find("title").string)) + "\n\n"
    outputString += "## " + removeWeirdSpacing(str(soup.find("author").string)) + "\n\n"

    outputString += transformSoupToMarkdown(documentList)

    # Supprime les faux espaces utilisés pour l'indentation
    outputString = outputString.replace(' ', '')

    # Ecriture dans le fichier
    with open(filename+'.md', 'w') as f:
        f.write(outputString)
        
if __name__ == '__main__':
    filename = "graphes (copia).tex"
    DIR = 'images/{}/'.format((filename).replace(' ', '_'))
    run(filename)



"""Parsing mechanisms should not be directly invoked publicly, as they are
subject to change."""

from .utils import Token, Buffer, MixedBuffer, CharToLineOffset
from .data import *
from .data import arg_type
from .tokens import (
    TC,
    tokenize,
    SKIP_ENV_NAMES,
    MATH_ENV_NAMES,
)
import functools
import string
import sys


MODE_MATH = 'mode:math'
MODE_NON_MATH = 'mode:non-math'
MATH_SIMPLE_ENVS = (
    TexDisplayMathModeEnv,
    TexMathModeEnv,
    TexDisplayMathEnv,
    TexMathEnv
)
MATH_TOKEN_TO_ENV = {env.token_begin: env for env in MATH_SIMPLE_ENVS}
ARG_BEGIN_TO_ENV = {arg.token_begin: arg for arg in arg_type}

SIGNATURES = {
    'def': (2, 0),
    'textbf': (1, 0),
    'section': (1, 1),
    'label': (1, 0),
    'cap': (0, 0),
    'cup': (0, 0),
    'in': (0, 0),
    'notin': (0, 0),
    'infty': (0, 0),
    'noindent': (0, 0),
}


__all__ = ['read_expr', 'read_tex']


def read_tex(buf, skip_envs=(), tolerance=0):
    r"""Parse all expressions in buffer

    :param Buffer buf: a buffer of tokens
    :param Tuple[str] skip_envs: environments to skip parsing
    :param int tolerance: error tolerance level (only supports 0 or 1)
    :return: iterable over parsed expressions
    :rtype: Iterable[TexExpr]
    """
    while buf.hasNext():
        yield read_expr(buf,
                        skip_envs=SKIP_ENV_NAMES + skip_envs,
                        tolerance=tolerance)


def make_read_peek(f):
    r"""Make any reader into a peek function.

    The wrapped function still parses the next sequence of tokens in the
    buffer but rolls back the buffer position afterwards.

    >>> from TexSoup.category import categorize
    >>> from TexSoup.tokens import tokenize
    >>> def read(buf):
    ...     buf.forward(3)
    >>> buf = Buffer(tokenize(categorize(r'\item testing \textbf{hah}')))
    >>> buf.position
    0
    >>> make_read_peek(read)(buf)
    >>> buf.position
    0
    """
    @functools.wraps(f)
    def wrapper(buf, *args, **kwargs):
        start = buf.position
        ret = f(buf, *args, **kwargs)
        buf.backward(buf.position - start)
        return ret
    return wrapper


def read_expr(src, skip_envs=(), tolerance=0, mode=MODE_NON_MATH):
    r"""Read next expression from buffer

    :param Buffer src: a buffer of tokens
    :param Tuple[str] skip_envs: environments to skip parsing
    :param int tolerance: error tolerance level (only supports 0 or 1)
    :param str mode: math or not math mode
    :return: parsed expression
    :rtype: [TexExpr, Token]
    """
    c = next(src)
    if c.category in MATH_TOKEN_TO_ENV.keys():
        expr = MATH_TOKEN_TO_ENV[c.category]([], position=c.position)
        return read_math_env(src, expr, tolerance=tolerance)
    elif c.category == TC.Escape:
        name, args = read_command(src, tolerance=tolerance, mode=mode)
        if name == 'item':
            assert mode != MODE_MATH, r'Command \item invalid in math mode.'
            contents = read_item(src)
            expr = TexCmd(name, contents, args, position=c.position)
        elif name == 'begin':
            assert args, 'Begin command must be followed by an env name.'
            expr = TexNamedEnv(
                args[0].string, args=args[1:], position=c.position)
            if expr.name in MATH_ENV_NAMES:
                mode = MODE_MATH
            if expr.name in skip_envs:
                read_skip_env(src, expr)
            else:
                read_env(src, expr, skip_envs=skip_envs,tolerance=tolerance, mode=mode)
        else:
            expr = TexCmd(name, args=args, position=c.position)
        return expr
    if c.category == TC.GroupBegin:
        return read_arg(src, c, tolerance=tolerance)

    assert isinstance(c, Token)
    return TexText(c)


################
# ENVIRONMENTS #
################


def read_item(src, tolerance=0):
    r"""Read the item content. Assumes escape has just been parsed.

    There can be any number of whitespace characters between \item and the
    first non-whitespace character. Any amount of whitespace between subsequent
    characters is also allowed.

    \item can also take an argument.

    :param Buffer src: a buffer of tokens
    :param int tolerance: error tolerance level (only supports 0 or 1)
    :return: contents of the item and any item arguments

    >>> from TexSoup.category import categorize
    >>> from TexSoup.tokens import tokenize
    >>> def read_item_from(string, skip=2):
    ...     buf = tokenize(categorize(string))
    ...     _ = buf.forward(skip)
    ...     return read_item(buf)
    >>> read_item_from(r'\item aaa {bbb} ccc\end{itemize}')
    [' aaa ', BraceGroup('bbb'), ' ccc']
    >>> read_item_from(r'\item aaa \textbf{itemize}\item no')
    [' aaa ', TexCmd('textbf', [BraceGroup('itemize')])]
    >>> read_item_from(r'\item WITCH [nuuu] DOCTORRRR 👩🏻‍⚕️')
    [' WITCH ', '[', 'nuuu', ']', ' DOCTORRRR 👩🏻‍⚕️']
    >>> read_item_from(r'''\begin{itemize}
    ... \item
    ... \item first item
    ... \end{itemize}''', skip=8)
    ['\n']
    >>> read_item_from(r'''\def\itemeqn{\item}''', skip=7)
    []
    """
    extras = []

    while src.hasNext():
        if src.peek().category == TC.Escape:
            cmd_name, _ = make_read_peek(read_command)(
                src, 1, skip=1, tolerance=tolerance)
            if cmd_name in ('end', 'item'):
                return extras
        elif src.peek().category == TC.GroupEnd:
            break
        extras.append(read_expr(src, tolerance=tolerance))
    return extras


def unclosed_env_handler(src, expr, end):
    """Handle unclosed environments.

    Currently raises an end-of-file error. In the future, this can be the hub
    for unclosed-environment fault tolerance.

    :param Buffer src: a buffer of tokens
    :param TexExpr expr: expression for the environment
    :param int tolerance: error tolerance level (only supports 0 or 1)
    :param end str: Actual end token (as opposed to expected)
    """
    clo = CharToLineOffset(str(src))
    explanation = 'Instead got %s' % end if end else 'Reached end of file.'
    line, offset = clo(src.position)
    raise EOFError('[Line: %d, Offset: %d] "%s" env expecting %s. %s' % (
        line, offset, expr.name, expr.end, explanation))


def read_math_env(src, expr, tolerance=0):
    r"""Read the environment from buffer.

    Advances the buffer until right after the end of the environment. Adds
    parsed content to the expression automatically.

    :param Buffer src: a buffer of tokens
    :param TexExpr expr: expression for the environment
    :rtype: TexExpr

    >>> from TexSoup.category import categorize
    >>> from TexSoup.tokens import tokenize
    >>> buf = tokenize(categorize(r'\min_x \|Xw-y\|_2^2'))
    >>> read_math_env(buf, TexMathModeEnv())
    Traceback (most recent call last):
        ...
    EOFError: [Line: 0, Offset: 7] "$" env expecting $. Reached end of file.
    """
    contents = []
    while src.hasNext() and src.peek().category != expr.token_end:
        contents.append(read_expr(src, tolerance=tolerance, mode=MODE_MATH))
    if not src.hasNext() or src.peek().category != expr.token_end:
        unclosed_env_handler(src, expr, src.peek())
    next(src)
    expr.append(*contents)
    return expr


def read_skip_env(src, expr):
    r"""Read the environment from buffer, WITHOUT parsing contents

    Advances the buffer until right after the end of the environment. Adds
    UNparsed content to the expression automatically.

    :param Buffer src: a buffer of tokens
    :param TexExpr expr: expression for the environment
    :rtype: TexExpr

    >>> from TexSoup.category import categorize
    >>> from TexSoup.tokens import tokenize
    >>> buf = tokenize(categorize(r' \textbf{aa \end{foobar}ha'))
    >>> read_skip_env(buf, TexNamedEnv('foobar'))
    TexNamedEnv('foobar', [' \\textbf{aa '], [])
    >>> buf = tokenize(categorize(r' \textbf{aa ha'))
    >>> read_skip_env(buf, TexNamedEnv('foobar'))  #doctest:+ELLIPSIS
    Traceback (most recent call last):
        ...
    EOFError: ...
    """
    def condition(s): return s.startswith('\\end{%s}' % expr.name)
    contents = [src.forward_until(condition, peek=False)]
    if not src.startswith('\\end{%s}' % expr.name):
        unclosed_env_handler(src, expr, src.peek((0, 6)))
    src.forward(5)
    expr.append(*contents)
    return expr


def read_env(src, expr, skip_envs=(), tolerance=0, mode=MODE_NON_MATH):
    r"""Read the environment from buffer.

    Advances the buffer until right after the end of the environment. Adds
    parsed content to the expression automatically.

    :param Buffer src: a buffer of tokens
    :param TexExpr expr: expression for the environment
    :param int tolerance: error tolerance level (only supports 0 or 1)
    :param str mode: math or not math mode
    :rtype: TexExpr

    >>> from TexSoup.category import categorize
    >>> from TexSoup.tokens import tokenize
    >>> buf = tokenize(categorize(' tingtang \\end\n{foobar}walla'))
    >>> read_env(buf, TexNamedEnv('foobar'))
    TexNamedEnv('foobar', [' tingtang '], [])
    >>> buf = tokenize(categorize(' tingtang \\end\n\n{foobar}walla'))
    >>> read_env(buf, TexNamedEnv('foobar')) #doctest: +ELLIPSIS
    Traceback (most recent call last):
        ...
    EOFError: [Line: 0, Offset: 1] ...
    >>> buf = tokenize(categorize(' tingtang \\end\n\n{nope}walla'))
    >>> read_env(buf, TexNamedEnv('foobar'), tolerance=1)  # error tolerance
    TexNamedEnv('foobar', [' tingtang '], [])
    """
    contents = []
    while src.hasNext():
        if src.peek().category == TC.Escape:
            name, args = make_read_peek(read_command)(
                src, skip=1, tolerance=tolerance, mode=mode)
            if name == 'end':
                break
        contents.append(read_expr(src, skip_envs=skip_envs, tolerance=tolerance, mode=mode))
    error = not src.hasNext() or not args or args[0].string != expr.name
    if error and tolerance == 0:
        unclosed_env_handler(src, expr, src.peek((0, 6)))
    elif not error:
        src.forward(5)
    expr.append(*contents)
    return expr


############
# COMMANDS #
############


# TODO: handle macro-weirdness e.g., \def\blah[#1][[[[[[[[#2{"#1 . #2"}
# TODO: add newcommand macro
def read_args(src, n_required=-1, n_optional=-1, args=None, tolerance=0,
        mode=MODE_NON_MATH):
    r"""Read all arguments from buffer.

    This function assumes that the command name has already been parsed. By
    default, LaTeX allows only up to 9 arguments of both types, optional
    and required. If `n_optional` is not set, all valid bracket groups are
    captured. If `n_required` is not set, all valid brace groups are
    captured.

    :param Buffer src: a buffer of tokens
    :param TexArgs args: existing arguments to extend
    :param int n_required: Number of required arguments. If < 0, all valid
                           brace groups will be captured.
    :param int n_optional: Number of optional arguments. If < 0, all valid
                           bracket groups will be captured.
    :param int tolerance: error tolerance level (only supports 0 or 1)
    :param str mode: math or not math mode
    :return: parsed arguments
    :rtype: TexArgs

    >>> from TexSoup.category import categorize
    >>> from TexSoup.tokens import tokenize
    >>> test = lambda s, *a, **k: read_args(tokenize(categorize(s)), *a, **k)
    >>> test('[walla]{walla}{ba]ng}')  # 'regular' arg parse
    [BracketGroup('walla'), BraceGroup('walla'), BraceGroup('ba', ']', 'ng')]
    >>> test('\t[wa]\n{lla}\n\n{b[ing}')  # interspersed spacers + 2 newlines
    [BracketGroup('wa'), BraceGroup('lla')]
    >>> test('\t[\t{a]}bs', 2, 0)  # use char as arg, since no opt args
    [BraceGroup('['), BraceGroup('a', ']')]
    >>> test('\n[hue]\t[\t{a]}', 2, 1)  # check stop opt arg capture
    [BracketGroup('hue'), BraceGroup('['), BraceGroup('a', ']')]
    >>> test('\t\\item')
    []
    >>> test('   \t    \n\t \n{bingbang}')
    []
    >>> test('[tempt]{ing}[WITCH]{doctorrrr}', 0, 0)
    []
    """
    args = args or TexArgs()
    if n_required == 0 and n_optional == 0:
        return args

    n_optional = read_arg_optional(src, args, n_optional, tolerance, mode)
    n_required = read_arg_required(src, args, n_required, tolerance, mode)

    if src.hasNext() and src.peek().category == TC.BracketBegin:
        n_optional = read_arg_optional(src, args, n_optional, tolerance, mode)
    if src.hasNext() and src.peek().category == TC.GroupBegin:
        n_required = read_arg_required(src, args, n_required, tolerance, mode)
    return args


def read_arg_optional(
        src, args, n_optional=-1, tolerance=0, mode=MODE_NON_MATH):
    """Read next optional argument from buffer.

    If the command has remaining optional arguments, look for:

       a. A spacer. Skip the spacer if it exists.
       b. A bracket delimiter. If the optional argument is bracket-delimited,
          the contents of the bracket group are used as the argument.

    :param Buffer src: a buffer of tokens
    :param TexArgs args: existing arguments to extend
    :param int n_optional: Number of optional arguments. If < 0, all valid
                           bracket groups will be captured.
    :param int tolerance: error tolerance level (only supports 0 or 1)
    :param str mode: math or not math mode
    :return: number of remaining optional arguments
    :rtype: int
    """
    while n_optional != 0:
        spacer = read_spacer(src)
        if not (src.hasNext() and src.peek().category == TC.BracketBegin):
            if spacer:
                src.backward(1)
            break
        args.append(read_arg(src, next(src), tolerance=tolerance, mode=mode))
        n_optional -= 1
    return n_optional


def read_arg_required(
        src, args, n_required=-1, tolerance=0, mode=MODE_NON_MATH):
    r"""Read next required argument from buffer.

    If the command has remaining required arguments, look for:

       a. A spacer. Skip the spacer if it exists.
       b. A curly-brace delimiter. If the required argument is brace-delimited,
          the contents of the brace group are used as the argument.
       c. Spacer or not, if a brace group is not found, simply use the next
          character, unless it is a backslash, in which case use the full command name

    :param Buffer src: a buffer of tokens
    :param TexArgs args: existing arguments to extend
    :param int n_required: Number of required arguments. If < 0, all valid
                           brace groups will be captured.
    :param int tolerance: error tolerance level (only supports 0 or 1)
    :param str mode: math or not math mode
    :return: number of remaining optional arguments
    :rtype: int

    >>> from TexSoup.category import categorize
    >>> from TexSoup.tokens import tokenize
    >>> buf = tokenize(categorize('{wal]la}\n{ba ng}\n'))
    >>> args = TexArgs()
    >>> read_arg_required(buf, args)  # 'regular' arg parse
    -3
    >>> args
    [BraceGroup('wal', ']', 'la'), BraceGroup('ba ng')]
    >>> buf.hasNext() and buf.peek().category == TC.MergedSpacer
    True
    """
    while n_required != 0 and src.hasNext():
        spacer = read_spacer(src)

        if src.hasNext() and src.peek().category == TC.GroupBegin:
            args.append(read_arg(
                src, next(src), tolerance=tolerance, mode=mode))
            n_required -= 1
            continue
        elif src.hasNext() and n_required > 0:
            next_token = next(src)
            if next_token.category == TC.Escape:
                name, _ = read_command(src, 0, 0, tolerance=tolerance, mode=mode)
                args.append(TexCmd(name, position=next_token.position))
            else:
                args.append('{%s}' % next_token)
            n_required -= 1
            continue

        if spacer:
            src.backward(1)
        break
    return n_required


def read_arg(src, c, tolerance=0, mode=MODE_NON_MATH):
    r"""Read the argument from buffer.

    Advances buffer until right before the end of the argument.

    :param Buffer src: a buffer of tokens
    :param str c: argument token (starting token)
    :param int tolerance: error tolerance level (only supports 0 or 1)
    :param str mode: math or not math mode
    :return: the parsed argument
    :rtype: TexGroup

    >>> from TexSoup.category import categorize
    >>> from TexSoup.tokens import tokenize
    >>> s = r'''{\item\abovedisplayskip=2pt\abovedisplayshortskip=0pt~\vspace*{-\baselineskip}}'''
    >>> buf = tokenize(categorize(s))
    >>> read_arg(buf, next(buf))
    BraceGroup(TexCmd('item'))
    >>> buf = tokenize(categorize(r'{\incomplete! [complete]'))
    >>> read_arg(buf, next(buf), tolerance=1)
    BraceGroup(TexCmd('incomplete'), '! ', '[', 'complete', ']')
    """
    content = [c]
    arg = ARG_BEGIN_TO_ENV[c.category]
    while src.hasNext():
        if src.peek().category == arg.token_end:
            src.forward()
            return arg(*content[1:], position=c.position)
        else:
            content.append(read_expr(src, tolerance=tolerance, mode=mode))

    if tolerance == 0:
        clo = CharToLineOffset(str(src))
        line, offset = clo(c.position)
        raise TypeError(
            '[Line: %d, Offset %d] Malformed argument. First and last elements '
            'must match a valid argument format. In this case, TexSoup'
            ' could not find matching punctuation for: %s.\n'
            'Just finished parsing: %s' %
            (line, offset, c, content))
    return arg(*content[1:], position=c.position)


def read_spacer(buf):
    r"""Extracts the next spacer, if there is one, before non-whitespace

    Define a spacer to be a contiguous string of only whitespace, with at most
    one line break.

    >>> from TexSoup.category import categorize
    >>> from TexSoup.tokens import tokenize
    >>> read_spacer(Buffer(tokenize(categorize('   \t    \n'))))
    '   \t    \n'
    >>> read_spacer(Buffer(tokenize(categorize('   \t    \n\t \n  \t\n'))))
    '   \t    \n\t '
    >>> read_spacer(Buffer(tokenize(categorize('{'))))
    ''
    >>> read_spacer(Buffer(tokenize(categorize('   \t    \na'))))
    ''
    >>> read_spacer(Buffer(tokenize(categorize('   \t    \n\t \n  \t\na'))))
    '   \t    \n\t '
    """
    if buf.hasNext() and buf.peek().category == TC.MergedSpacer:
        return next(buf)
    return ''


def read_command(buf, n_required_args=-1, n_optional_args=-1, skip=0,
                 tolerance=0, mode=MODE_NON_MATH):
    r"""Parses command and all arguments. Assumes escape has just been parsed.

    No whitespace is allowed between escape and command name. e.g.,
    :code:`\ textbf` is a backslash command, then text :code:`textbf`. Only
    :code:`\textbf` is the bold command.

    >>> from TexSoup.category import categorize
    >>> from TexSoup.tokens import tokenize
    >>> buf = Buffer(tokenize(categorize('\\sect  \t    \n\t{wallawalla}')))
    >>> next(buf)
    '\\'
    >>> read_command(buf)
    ('sect', [BraceGroup('wallawalla')])
    >>> buf = Buffer(tokenize(categorize('\\sect  \t   \n\t \n{bingbang}')))
    >>> _ = next(buf)
    >>> read_command(buf)
    ('sect', [])
    >>> buf = Buffer(tokenize(categorize('\\sect{ooheeeee}')))
    >>> _ = next(buf)
    >>> read_command(buf)
    ('sect', [BraceGroup('ooheeeee')])
    >>> buf = Buffer(tokenize(categorize(r'\item aaa {bbb} ccc\end{itemize}')))
    >>> read_command(buf, skip=1)
    ('item', [])
    >>> buf.peek()
    ' aaa '

    # >>> buf = Buffer(tokenize(categorize('\\sect abcd')))
    # >>> _ = next(buf)
    # >>> read_command(buf)
    # ('sect', ('a',))
    """
    for _ in range(skip):
        next(buf)

    name = next(buf)
    token = Token('', buf.position)
    if n_required_args < 0 and n_optional_args < 0:
        n_required_args, n_optional_args = SIGNATURES.get(name, (-1, -1))
    args = read_args(buf, n_required_args, n_optional_args,
                     tolerance=tolerance, mode=mode)
    return name, args



"""Categorize all characters into one of category codes."""

from .utils import CC, Token, to_buffer
import string


# Core category codes
# https://www.overleaf.com/learn/latex/Table_of_TeX_category_codes
others = set(string.printable) - set(string.ascii_letters) - \
    set('{}\\$&\n\r#^_~%\x00\x7d \t[]()')
CATEGORY_CODES = {
    CC.Escape:      '\\',
    CC.GroupBegin:  '{',
    CC.GroupEnd:    '}',
    CC.MathSwitch:  '$',
    CC.Alignment:   '&',  # not used
    CC.EndOfLine:   ('\n', '\r'),
    CC.Macro:       '#',  # not used
    CC.Superscript: '^',  # not used
    CC.Subscript:   '_',  # not used
    CC.Ignored:     chr(0),
    CC.Spacer:      (chr(32), chr(9)),
    CC.Letter:      tuple(string.ascii_letters),  # + lots of unicode
    CC.Other:       tuple(others),
    CC.Active:      '~',  # not used
    CC.Comment:     '%',
    CC.Invalid:      chr(127),

    # custom
    CC.BracketBegin: '[',
    CC.BracketEnd:  ']',
    CC.ParenBegin:  '(',
    CC.ParenEnd:    ')'
}


@to_buffer()
def categorize(text):
    r"""Generator for category code tokens on text, ignoring comments.

    :param Union[str,iterator,Buffer] text: LaTeX to process

    >>> chars = list(categorize(r'\bf{}%[ello+😂'))
    >>> chars[0].category
    <CategoryCodes.Escape: 1>
    >>> chars[1].category
    <CategoryCodes.Letter: 12>
    >>> chars[3].category
    <CategoryCodes.GroupBegin: 2>
    >>> chars[4].category
    <CategoryCodes.GroupEnd: 3>
    >>> chars[5].category
    <CategoryCodes.Comment: 15>
    >>> chars[6].category
    <CategoryCodes.BracketBegin: 19>
    >>> chars[-2].category
    <CategoryCodes.Other: 13>
    >>> chars[-1].category
    <CategoryCodes.Other: 13>
    >>> print(*chars)
    \ b f { } % [ e l l o + 😂
    >>> next(categorize(r'''
    ... ''')).category
    <CategoryCodes.EndOfLine: 6>
    """
    for position, char in enumerate(text):

        value = None
        for cc, values in CATEGORY_CODES.items():
            if char in values:
                value = char
                break

        if value is None:
            yield Token(char, position, CC.Other)
        else:
            yield Token(char, position, cc)



disponible_environments=[
    "align"
    "align*"
    "alignat"
    "alignat*"
    "aligned"
    "alignedat"
    "array"
    "bmatrix"
    "Bmatrix"
    "bmatrix*"
    "Bmatrix*"
    "bsmallmatrix"
    "Bsmallmatrix"
    "bsmallmatrix*"
    "Bsmallmatrix*"
    "cases"
    "cases*"
    "CD"
    "crampedsubarray"
    "dcases"
    "dcases*"
    "drcases"
    "drcases*"
    "empheq"
    "eqnarray"
    "eqnarray*"
    "equation"
    "equation*"
    "flalign"
    "flalign*"
    "gather"
    "gather*"
    "gathered"
    "lgathered"
    "matrix"
    "matrix*"
    "multline"
    "multline*"
    "multlined"
    "numcases"
    "pmatrix"
    "pmatrix*"
    "prooftree"
    "psmallmatrix"
    "psmallmatrix*"
    "rcases"
    "rcases*"
    "rgathered"
    "smallmatrix"
    "smallmatrix*"
    "split"
    "spreadlines"
    "subarray"
    "subnumcases"
    "vmatrix"
    "Vmatrix"
    "vmatrix*"
    "Vmatrix*"
    "vsmallmatrix"
    "Vsmallmatrix"
    "vsmallmatrix*"
    "Vsmallmatrix*"
    "xalignat"
    "xalignat*"
    "xxalignat"
]

skip_envs = ()


from .reader import read_expr, read_tex
from .data import *
from .utils import *
from .tokens import tokenize
from .category import categorize
import itertools


def read(tex, skip_envs=(), tolerance=0):
    """Read and parse all LaTeX source.

    :param Union[str,iterable] tex: LaTeX source
    :param Union[str] skip_envs: names of environments to skip parsing
    :param int tolerance: error tolerance level (only supports 0 or 1)
    :return TexEnv: the global environment
    """
    if not isinstance(tex, str):
        tex = ''.join(itertools.chain(*tex))
    buf = categorize(tex)
    buf = tokenize(buf)
    buf = read_tex(buf, skip_envs=skip_envs, tolerance=tolerance)
    return TexEnv('[tex]', begin='', end='', contents=buf), tex
